"""Tests for unread messages detection."""

from unittest.mock import AsyncMock, patch

import pytest

from wilhelmina.client import WilmaClient, WilmaError


@pytest.mark.asyncio
@patch("wilhelmina.client._has_playwright", True)
@patch("wilhelmina.client.async_playwright")
async def test_get_unread_message_ids(mock_playwright) -> None:
    """Test getting unread message IDs with Playwright."""
    # Initialize client
    client = WilmaClient("https://test.inschool.fi")
    client.user_id = "!12345"
    client._sid = "test_sid"
    client.headless = True

    # Setup Playwright mocks
    mock_browser = AsyncMock()
    mock_context = AsyncMock()
    mock_page = AsyncMock()
    mock_response = AsyncMock()

    # Setup the mock playwright structure
    mock_playwright_instance = AsyncMock()
    mock_playwright.return_value.__aenter__.return_value = mock_playwright_instance
    mock_playwright_instance.chromium.launch.return_value = mock_browser
    mock_browser.new_context.return_value = mock_context
    mock_context.new_page.return_value = mock_page

    # Setup the page goto response
    mock_response.status = 200
    mock_page.goto.return_value = mock_response

    # Create a sample HTML response with message rows
    html_content = """
    <!DOCTYPE html>
    <html>
    <body>
        <table id="message-list-table">
            <tbody>
                <tr>
                    <td></td>
                    <td>Unread</td>
                    <td></td>
                    <td></td>
                    <td><input type="checkbox" name="mid" value="123"></td>
                </tr>
                <tr>
                    <td></td>
                    <td></td>
                    <td></td>
                    <td></td>
                    <td><input type="checkbox" name="mid" value="456"></td>
                </tr>
                <tr>
                    <td></td>
                    <td>Unread</td>
                    <td></td>
                    <td></td>
                    <td><input type="checkbox" name="mid" value="789"></td>
                </tr>
            </tbody>
        </table>
    </body>
    </html>
    """
    mock_page.content.return_value = html_content

    # Call the method
    unread_message_ids = await client._get_unread_message_ids()

    # Verify the results
    assert unread_message_ids == {123, 789}

    # Verify the correct methods were called
    mock_playwright_instance.chromium.launch.assert_called_once_with(headless=True)
    mock_browser.new_context.assert_called_once()
    mock_context.new_page.assert_called_once()
    mock_page.goto.assert_called_once_with("https://test.inschool.fi/!12345/messages")
    mock_page.wait_for_selector.assert_called_once_with(
        "table#message-list-table", state="attached", timeout=10000
    )
    mock_page.content.assert_called_once()
    mock_browser.close.assert_called_once()


@pytest.mark.asyncio
@patch("wilhelmina.client._has_playwright", True)
@patch("wilhelmina.client.async_playwright")
async def test_get_unread_message_ids_failure(mock_playwright) -> None:
    """Test error handling in getting unread message IDs."""
    # Initialize client
    client = WilmaClient("https://test.inschool.fi")
    client.user_id = "!12345"
    client._sid = "test_sid"

    # Setup Playwright mocks
    mock_browser = AsyncMock()
    mock_context = AsyncMock()
    mock_page = AsyncMock()
    mock_response = AsyncMock()

    # Setup the mock playwright structure
    mock_playwright_instance = AsyncMock()
    mock_playwright.return_value.__aenter__.return_value = mock_playwright_instance
    mock_playwright_instance.chromium.launch.return_value = mock_browser
    mock_browser.new_context.return_value = mock_context
    mock_context.new_page.return_value = mock_page

    # Setup the page goto response with error status
    mock_response.status = 404  # Not found
    mock_page.goto.return_value = mock_response

    # Call the method, should raise an error
    with pytest.raises(WilmaError, match="Failed to load messages page"):
        await client._get_unread_message_ids()

    # Verify the browser was closed
    mock_browser.close.assert_called_once()


@pytest.mark.asyncio
@patch("wilhelmina.client._has_playwright", False)
async def test_get_unread_message_ids_no_playwright() -> None:
    """Test behavior when Playwright is not available."""
    # Initialize client
    client = WilmaClient("https://test.inschool.fi")
    client.user_id = "!12345"
    client._sid = "test_sid"

    # With _has_playwright=False, should return an empty set without errors
    unread_message_ids = await client._get_unread_message_ids()
    assert unread_message_ids == set()


@pytest.mark.asyncio
@patch("wilhelmina.client._has_playwright", True)
@patch("wilhelmina.client.async_playwright")
async def test_get_unread_message_ids_invalid_values(mock_playwright) -> None:
    """Test handling of invalid message IDs."""
    # Initialize client
    client = WilmaClient("https://test.inschool.fi")
    client.user_id = "!12345"
    client._sid = "test_sid"

    # Setup Playwright mocks
    mock_browser = AsyncMock()
    mock_context = AsyncMock()
    mock_page = AsyncMock()
    mock_response = AsyncMock()

    # Setup the mock playwright structure
    mock_playwright_instance = AsyncMock()
    mock_playwright.return_value.__aenter__.return_value = mock_playwright_instance
    mock_playwright_instance.chromium.launch.return_value = mock_browser
    mock_browser.new_context.return_value = mock_context
    mock_context.new_page.return_value = mock_page

    # Setup the page goto response
    mock_response.status = 200
    mock_page.goto.return_value = mock_response

    # Create a sample HTML response with invalid message IDs
    html_content = """
    <!DOCTYPE html>
    <html>
    <body>
        <table id="message-list-table">
            <tbody>
                <tr>
                    <td></td>
                    <td>Unread</td>
                    <td></td>
                    <td></td>
                    <td><input type="checkbox" name="mid" value="invalid"></td>
                </tr>
                <tr>
                    <td></td>
                    <td>Unread</td>
                    <td></td>
                    <td></td>
                    <td><input type="checkbox" name="mid"></td>
                </tr>
                <tr>
                    <td></td>
                    <td>Unread</td>
                    <td></td>
                    <td></td>
                    <td></td>
                </tr>
            </tbody>
        </table>
    </body>
    </html>
    """
    mock_page.content.return_value = html_content

    # Call the method - should handle invalid IDs gracefully
    unread_message_ids = await client._get_unread_message_ids()

    # Should return an empty set since all IDs were invalid
    assert unread_message_ids == set()
    mock_browser.close.assert_called_once()



"""Wilma API client."""

import datetime
import importlib.util
import logging
from typing import Any, Optional, Set
from urllib.parse import urlparse

import aiohttp
from aiohttp import ClientResponse
from bs4 import BeautifulSoup

from wilhelmina.models import Message

# Allow for running without Playwright
_has_playwright = importlib.util.find_spec("playwright") is not None
if _has_playwright:
    from playwright._impl._api_structures import SetCookieParam
    from playwright.async_api import async_playwright

logger = logging.getLogger(__name__)


class WilmaError(Exception):
    """Base exception for Wilma API errors."""


class AuthenticationError(WilmaError):
    """Authentication error."""


class WilmaClient:
    """Client for Wilma API."""

    def __init__(
        self,
        base_url: str,
        session: Optional[aiohttp.ClientSession] = None,
        debug: bool = False,
        headless: bool = True,
    ) -> None:
        """Initialize the client."""
        self.base_url = base_url.rstrip("/")
        self.session = session
        self.user_id: Optional[str] = None
        self._sid: Optional[str] = None
        self.debug = debug
        self.headless = headless
        self._owns_session = session is None
        self.username = None
        self.password = None

        if debug:
            logger.setLevel(logging.DEBUG)

    async def _ensure_session(self) -> aiohttp.ClientSession:
        """Ensure that a session exists."""
        if self.session is None or self.session.closed:
            self.session = aiohttp.ClientSession()
            self._owns_session = True
        return self.session

    async def close(self) -> None:
        """Close the HTTP session if we created it."""
        if self._owns_session and self.session and not self.session.closed:
            await self.session.close()
            self.session = None

    async def __aenter__(self) -> "WilmaClient":
        await self._ensure_session()
        return self

    async def __aexit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        await self.close()

    def _check_auth(self) -> None:
        """Check if authenticated and raise error if not."""
        if not self.user_id or not self._sid:
            logger.debug("Not authenticated")
            raise AuthenticationError("Not authenticated")

    async def login(self, username: str | None = None, password: str | None = None) -> None:
        """Login to Wilma."""
        self.username = username or self.username
        self.password = password or self.password

        if not self.username or not self.password:
            raise AuthenticationError("Username and password must be provided")

        session = await self._ensure_session()
        logger.debug("Logging in to %s as %s", self.base_url, self.username)

        # Get SESSIONID token
        async with session.get(f"{self.base_url}/token") as response:
            if response.status != 200:
                raise AuthenticationError(f"Failed to get token: {response.status}")

            login_id = response.cookies.get("Wilma2LoginID")
            if not login_id:
                raise AuthenticationError("No login ID cookie found")

            session_id = login_id.value
            logger.debug("Got session ID: %s", session_id)

        # Perform login
        data = {
            "Login": self.username,
            "Password": self.password,
            "SESSIONID": session_id,
        }

        async with session.post(
            f"{self.base_url}/login", data=data, allow_redirects=False
        ) as response:
            if response.status != 302 and response.status != 303:
                raise AuthenticationError(f"Login failed: {response.status}")

            sid = response.cookies.get("Wilma2SID")
            if not sid:
                raise AuthenticationError("No SID cookie found")

            self._sid = sid.value

            location = response.headers.get("Location")
            if not location:
                raise AuthenticationError("No Location header in response")

            if "!" not in location:
                raise AuthenticationError(f"Invalid Location header: {location}")

            self.user_id = "!" + location.split("!")[1].split("?")[0]
            logger.debug("Logged in as %s", self.user_id)

    async def _get_unread_message_ids(self) -> Set[int]:
        """Get IDs of unread messages by parsing the HTML page with Playwright."""
        self._check_auth()

        if not _has_playwright:
            logger.debug("Playwright not available, skipping unread detection")
            return set()

        url = f"{self.base_url}/{self.user_id}/messages"
        unread_ids = set()

        async with async_playwright() as p:
            browser = await p.chromium.launch(headless=self.headless)
            context = await browser.new_context()

            # Set auth cookie with proper typing
            cookie_domain = urlparse(self.base_url).netloc
            if _has_playwright:
                cookie: SetCookieParam = {
                    "name": "Wilma2SID",
                    "value": self._sid or "",
                    "domain": cookie_domain,
                    "path": "/",
                    "sameSite": "Lax",
                }
                await context.add_cookies([cookie])

            page = await context.new_page()

            try:
                response = await page.goto(url)
                if not response or response.status != 200:
                    status = response.status if response else "No response"
                    raise WilmaError(f"Failed to load messages page: {status}")

                await page.wait_for_selector(
                    "table#message-list-table", state="attached", timeout=10000
                )
                html_content = await page.content()
                soup = BeautifulSoup(html_content, "html.parser")

                # Extract unread message IDs
                for row in soup.select("table#message-list-table tbody tr"):
                    checkbox = row.select_one('input[name="mid"]')
                    if checkbox and checkbox.get("value"):
                        value = checkbox.get("value")
                        if not isinstance(value, str):
                            continue
                        try:
                            message_id = int(value)
                            td = row.select_one("td:nth-child(2)")
                            if td and td.text.strip():
                                unread_ids.add(message_id)
                        except (ValueError, AttributeError):
                            continue

                logger.debug("Found %s unread messages", len(unread_ids))
            finally:
                await browser.close()

        return unread_ids

    async def _authenticated_request(
        self,
        url_template: str,
        method: str = "GET",
        data: dict[str, Any] | None = None,
        params: dict[str, str | int | bool] | None = None,
        retry: bool = False,
        **kwargs: Any,
    ) -> aiohttp.ClientResponse:
        """Make an authenticated request to the Wilma API."""
        self._check_auth()
        session = await self._ensure_session()

        # Replace {user_id} in template if present
        path = url_template
        if "{user_id}" in url_template and self.user_id:
            path = url_template.format(user_id=self.user_id)

        # Construct full URL with base_url
        url = f"{self.base_url}/{path.lstrip('/')}"

        # Add auth header
        headers = kwargs.pop("headers", {})
        headers["Wilma2SID"] = self._sid or ""

        # Make request
        method = method.upper()
        request_kwargs = {"headers": headers, **kwargs}

        if params:
            request_kwargs["params"] = params

        if data and method == "POST":
            request_kwargs["data"] = data

        logger.debug("Making %s request to %s", method, url)

        response: ClientResponse = await getattr(session, method.lower())(url, **request_kwargs)

        # Session expires after a while, and at times we get 403 errors, re-auth in these cases
        # but just do it once, since if a re-auth does not work, then we most likely are locked
        # our for a valid reason.
        reauth_matches = ["sessionexpired", "invalidsession"]
        realurl = str(response.request_info.real_url)

        if not retry and any(match in realurl for match in reauth_matches) or response.status == 403:
            logger.debug("Session expired, logging in again")
            await self.login()
            return await self._authenticated_request(url_template, method, data, params, True, **kwargs)

        if response.status >= 400:
            raise WilmaError(f"Request failed with status {response.status}: {url}")

        return response

    async def get_messages(
        self,
        only_unread: bool = False,
        after: datetime.datetime | None = None,
        with_content: bool = False,
        no_message_content_fetch_limit: bool = False,
    ) -> list[Message]:
        """Get messages list."""
        # Get unread message IDs
        unread_message_ids = await self._get_unread_message_ids()

        response = await self._authenticated_request("{user_id}/messages/list")
        data = await response.json()
        messages = []

        # Convert after to naive datetime if it has timezone info
        # We don't have any TZ info for the Wilma messages
        if after and after.tzinfo is not None:
            after = after.replace(tzinfo=None)

        for msg in data.get("Messages", []):
            message = Message.from_dict(msg)
            message.unread = message.id in unread_message_ids

            # Skip if only_unread is True and message is not unread
            if only_unread and not message.unread:
                continue

            # Skip if before is provided and message is not before the date
            if after and message.timestamp and message.format_timestamp() <= after:
                continue

            messages.append(message)

        # Fetch full content for messages if requested
        if with_content:
            if len(messages) > 10:
                if not no_message_content_fetch_limit:
                    raise WilmaError(
                        "Not allowed to fetch full message content for more than 10 messages."
                        "Set no_message_content_fetch_limit=True to override this limit."
                    )
                logger.warning(
                    "Fetching full message content for %s messages may take a while..",
                    len(messages),
                )
            for message in messages:
                try:
                    full_message = await self.get_message_content(message.id)
                    # Update message with content from full_message
                    message.content_html = full_message.content_html
                    # Preserve other attributes that might be in the full message
                    for attr, value in vars(full_message).items():
                        if attr != "unread" and getattr(message, attr, None) is None:
                            setattr(message, attr, value)
                except Exception as e:
                    logger.warning("Failed to fetch content for message %s: %s", message.id, e)

        return messages

    async def get_message_content(self, message_id: int) -> Message:
        """Get content of a message."""
        # Check if message is unread
        unread_message_ids = await self._get_unread_message_ids()

        url_template = "{user_id}/messages/{message_id}"
        url = url_template.format(user_id="{user_id}", message_id=message_id)

        response = await self._authenticated_request(url, params={"format": "json"})

        data = await response.json()
        message = Message.from_dict(data["messages"][0])
        message.unread = message_id in unread_message_ids

        return message



"""Tests for package __init__."""

import subprocess
from unittest.mock import MagicMock, patch

import pytest

from wilhelmina import __all__, install_playwright_deps


def test_all_exports() -> None:
    """Test that __all__ contains expected exports."""
    expected = [
        "WilmaClient",
        "WilmaError",
        "AuthenticationError",
        "Message",
        "MessagesList",
        "Sender",
        "install_playwright_deps",
    ]

    for item in expected:
        assert item in __all__


@pytest.mark.parametrize(
    "playwright_exists,force,quiet,expected_return",
    [
        (True, False, False, True),  # Normal case when playwright exists
        (False, False, False, False),  # Skip when playwright doesn't exist
        (False, True, False, False),  # Force but playwright doesn't exist (should still attempt)
        (True, True, True, True),  # Force and quiet
    ],
)
def test_install_playwright_deps(
    playwright_exists: bool, force: bool, quiet: bool, expected_return: bool
) -> None:
    """Test playwright dependency installation."""

    # Mock importlib.util.find_spec
    mock_find_spec = MagicMock()
    mock_find_spec.return_value = MagicMock() if playwright_exists else None

    # Mock subprocess.run
    mock_run = MagicMock()
    if not playwright_exists and not force:
        mock_run.side_effect = None  # Should not be called
    else:
        # When force=True but playwright doesn't exist, simulate failure
        if not playwright_exists and force:
            mock_run.side_effect = subprocess.SubprocessError("Command failed")
        else:
            mock_run.return_value = MagicMock()

    with (
        patch("wilhelmina.__init__.importlib.util.find_spec", mock_find_spec),
        patch("wilhelmina.__init__.subprocess.run", mock_run),
    ):
        result = install_playwright_deps(force=force, quiet=quiet)

        assert result == expected_return

        if not playwright_exists and not force:
            mock_run.assert_not_called()
        else:
            mock_run.assert_called_once()



"""Tests for utils."""

from wilhelmina.utils import html_to_markdown


def test_html_to_markdown() -> None:
    """Test conversion of HTML to Markdown."""
    # Test basic conversion
    html = "<p>This is a paragraph</p>"
    assert html_to_markdown(html) == "This is a paragraph"

    # Test with links
    html = "<p>Check out <a href='https://example.com'>this link</a></p>"
    assert html_to_markdown(html) == "Check out [this link](<https://example.com>)"

    # Test with formatting
    html = "<p><strong>Bold</strong> and <em>italic</em> text</p>"
    assert html_to_markdown(html) == "**Bold** and _italic_ text"

    # Test with lists
    html = "<ul><li>Item 1</li><li>Item 2</li></ul>"
    expected = "* Item 1\n  * Item 2"
    assert html_to_markdown(html) == expected

    # Test with empty string
    assert html_to_markdown("") == ""



"""Message models for Wilma API."""

from dataclasses import dataclass, field
from datetime import datetime
from typing import Any, List, Optional

from mashumaro import DataClassDictMixin
from mashumaro.config import BaseConfig

from wilhelmina.utils import html_to_markdown


@dataclass
class Sender(DataClassDictMixin):
    """Sender of a message."""

    name: str
    href: Optional[str] = None

    class Config(BaseConfig):
        aliases = {"name": "Name", "href": "Href"}


@dataclass
class Message(DataClassDictMixin):
    """A message in Wilma."""

    id: int
    subject: str
    timestamp: str
    folder: str
    sender_id: int
    sender_type: int
    sender: str
    content_html: str | None = None
    allow_forward: bool = True
    allow_reply: bool = True
    reply_list: List[Any] = field(default_factory=list)
    senders: List[Sender] = field(default_factory=list)
    unread: bool = False

    class Config(BaseConfig):
        aliases = {
            "id": "Id",
            "subject": "Subject",
            "timestamp": "TimeStamp",
            "folder": "Folder",
            "sender_id": "SenderId",
            "sender_type": "SenderType",
            "sender": "Sender",
            "content_html": "ContentHtml",
            "allow_forward": "AllowForward",
            "allow_reply": "AllowReply",
            "reply_list": "ReplyList",
            "senders": "Senders",
            # unread is not in the API response, we set it manually
        }

    def format_timestamp(self) -> datetime:
        """Format the timestamp string to a datetime object."""
        return datetime.strptime(self.timestamp, "%Y-%m-%d %H:%M")

    @property
    def content_markdown(self) -> str:
        """Get message content as markdown."""
        if self.content_html is None:
            raise ValueError("Message content is not available")
        return html_to_markdown(self.content_html)


@dataclass
class MessagesList:
    """List of messages response."""

    messages: List[Message] = field(default_factory=list)



"""Tests for models."""

from datetime import datetime

import pytest

from wilhelmina.models import Message, MessagesList, Sender


def test_sender_model() -> None:
    """Test Sender model."""
    data = {"name": "Teacher 1", "href": "/profiles/123"}
    sender = Sender(**data)

    assert sender.name == "Teacher 1"
    assert sender.href == "/profiles/123"


def test_message_model() -> None:
    """Test Message model."""
    data = {
        "id": 1,
        "subject": "Test message",
        "timestamp": "2025-04-16 13:31",
        "folder": "Inbox",
        "sender_id": 123,
        "sender_type": 1,
        "sender": "Teacher 1",
        "content_html": "<p>This is a test message.</p>",
        "allow_forward": True,
        "allow_reply": True,
        "reply_list": [],
        "senders": [Sender(name="Teacher 1", href="/profiles/123")],
        "unread": False,
    }

    message = Message(**data)

    assert message.id == 1
    assert message.subject == "Test message"
    assert message.timestamp == "2025-04-16 13:31"
    assert message.folder == "Inbox"
    assert message.sender_id == 123
    assert message.sender_type == 1
    assert message.sender == "Teacher 1"
    assert message.content_html == "<p>This is a test message.</p>"
    assert message.content_markdown == "This is a test message."
    assert message.allow_forward is True
    assert message.allow_reply is True
    assert message.reply_list == []
    assert len(message.senders) == 1
    assert message.senders[0].name == "Teacher 1"
    assert message.senders[0].href == "/profiles/123"


def test_content_markdown_error() -> None:
    """Test that accessing content_markdown raises error when content_html is None."""
    # Create message with None content_html
    message = Message(
        id=1,
        subject="Test message",
        timestamp="2025-04-16 13:31",
        folder="Inbox",
        sender_id=123,
        sender_type=1,
        sender="Teacher 1",
        content_html=None,
    )

    # Verify ValueError is raised when trying to access content_markdown
    with pytest.raises(ValueError, match="Message content is not available"):
        _ = message.content_markdown


def test_message_format_timestamp() -> None:
    """Test the format_timestamp method."""
    message = Message(
        id=1,
        subject="Test message",
        timestamp="2025-04-16 13:31",
        folder="Inbox",
        sender_id=123,
        sender_type=1,
        sender="Teacher 1",
    )

    # Test format_timestamp
    formatted = message.format_timestamp()
    assert isinstance(formatted, datetime)
    assert formatted.year == 2025
    assert formatted.month == 4
    assert formatted.day == 16
    assert formatted.hour == 13
    assert formatted.minute == 31


def test_messages_list_model() -> None:
    """Test MessagesList model."""
    messages = [
        Message(
            id=1,
            subject="Test message 1",
            timestamp="2025-04-16 13:31",
            folder="Inbox",
            sender_id=123,
            sender_type=1,
            sender="Teacher 1",
            content_html="<p>Message 1 content</p>",
            allow_forward=True,
            allow_reply=True,
            reply_list=[],
            senders=[Sender(name="Teacher 1", href="/profiles/123")],
            unread=True,
        ),
        Message(
            id=2,
            subject="Test message 2",
            timestamp="2025-04-15 10:20",
            folder="Inbox",
            sender_id=456,
            sender_type=1,
            sender="Teacher 2",
            content_html="<p>Message 2 content</p>",
            allow_forward=True,
            allow_reply=True,
            reply_list=[],
            senders=[],
            unread=False,
        ),
    ]

    messages_list = MessagesList(messages=messages)

    assert len(messages_list.messages) == 2
    assert messages_list.messages[0].id == 1
    assert messages_list.messages[0].subject == "Test message 1"
    assert messages_list.messages[1].id == 2
    assert messages_list.messages[1].subject == "Test message 2"



"""Test configuration."""

import datetime
import logging
from typing import Any, Dict, Generator, List
from unittest.mock import AsyncMock

import pytest

from wilhelmina.client import WilmaClient


def pytest_configure(config):
    """Configure pytest."""
    config.addinivalue_line("markers", "asyncio: mark test as an asyncio test")


@pytest.fixture(autouse=True)
def disable_logging() -> Generator[None, None, None]:
    """Disable logging for tests."""
    logging.disable(logging.CRITICAL)
    yield
    logging.disable(logging.NOTSET)


@pytest.fixture
def authenticated_client() -> WilmaClient:
    """Create an authenticated client instance."""
    client = WilmaClient("https://test.inschool.fi", debug=False)
    client.user_id = "!12345"
    client._sid = "test_sid"
    return client


@pytest.fixture
def messages_data() -> Dict[str, List[Dict[str, Any]]]:
    """Mock messages data."""
    return {
        "Messages": [
            {
                "Id": 1,
                "Subject": "Test message 1",
                "TimeStamp": "2025-04-16 13:31",
                "Folder": "Inbox",
                "SenderId": 123,
                "SenderType": 1,
                "Sender": "Teacher 1",
                "ContentHtml": "<p>Message 1 content</p>",
                "AllowForward": True,
                "AllowReply": True,
                "ReplyList": [],
                "Senders": [{"Name": "Teacher 1", "Href": "/profiles/123"}],
            },
            {
                "Id": 2,
                "Subject": "Test message 2",
                "TimeStamp": "2025-04-15 10:20",
                "Folder": "Inbox",
                "SenderId": 456,
                "SenderType": 1,
                "Sender": "Teacher 2",
                "ContentHtml": "<p>Message 2 content</p>",
                "AllowForward": True,
                "AllowReply": True,
                "ReplyList": [],
                "Senders": [{"Name": "Teacher 2", "Href": "/profiles/456"}],
            },
            {
                "Id": 3,
                "Subject": "Test message 3",
                "TimeStamp": "2025-04-10 09:15",  # Oldest
                "Folder": "Inbox",
                "SenderId": 789,
                "SenderType": 1,
                "Sender": "Teacher 3",
                "ContentHtml": "<p>Message 3 content</p>",
                "AllowForward": True,
                "AllowReply": True,
                "ReplyList": [],
                "Senders": [{"Name": "Teacher 3", "Href": "/profiles/789"}],
            },
        ]
    }


@pytest.fixture
def message_content_data() -> Dict[str, Any]:
    """Mock message content data."""
    return {
        "messages": [
            {
                "Id": 1,
                "Subject": "Test message 1",
                "TimeStamp": "2025-04-16 13:31",
                "Folder": "Inbox",
                "SenderId": 123,
                "SenderType": 1,
                "Sender": "Teacher 1",
                "ContentHtml": "<p>This is a test message content.</p>",
                "AllowForward": True,
                "AllowReply": True,
                "ReplyList": [],
                "Senders": [{"Name": "Teacher 1", "Href": "/profiles/123"}],
            }
        ]
    }


@pytest.fixture
def mock_authenticated_request(authenticated_client):
    """Mock the _authenticated_request method."""

    async def _mock_request(url_template, *args, **kwargs):
        resp = AsyncMock()
        resp.status = 200
        return resp

    authenticated_client._authenticated_request = AsyncMock(side_effect=_mock_request)
    return authenticated_client


@pytest.fixture
def date_filters():
    """Return a set of dates for filtering tests."""
    return {
        "before_all": datetime.datetime(2025, 4, 5),
        "middle": datetime.datetime(2025, 4, 12),
        "after_all": datetime.datetime(2025, 4, 25),
        "with_timezone": datetime.datetime(2025, 4, 12, tzinfo=datetime.timezone.utc),
    }



"""Tests for WilmaClient."""

from unittest.mock import AsyncMock, MagicMock, patch

import aiohttp
import pytest
from aiohttp import ClientResponse

from wilhelmina.client import AuthenticationError, WilmaClient, WilmaError
from wilhelmina.models import Message

################################
# Basic Client Functionality
################################


@pytest.mark.asyncio
async def test_init() -> None:
    """Test client initialization."""
    # Test with default parameters (no session provided)
    client = WilmaClient("https://test.inschool.fi", debug=True, headless=False)
    assert client.base_url == "https://test.inschool.fi"
    assert client.debug is True
    assert client.headless is False
    assert client.session is None
    assert client._owns_session is True

    # Make sure session is created by _ensure_session
    session = await client._ensure_session()
    assert session is client.session
    assert isinstance(session, aiohttp.ClientSession)

    # Close client
    await client.close()
    assert client.session is None

    # Test with provided session
    existing_session = aiohttp.ClientSession()
    client_with_session = WilmaClient("https://test.inschool.fi", session=existing_session)
    assert client_with_session.session is existing_session
    assert client_with_session._owns_session is False

    # Close should not close the session
    await client_with_session.close()
    assert not existing_session.closed

    # Clean up
    await existing_session.close()


@pytest.mark.asyncio
async def test_context_manager() -> None:
    """Test client as a context manager."""
    # We need to use a real session for this test
    async with WilmaClient("https://test.inschool.fi") as client:
        assert isinstance(client.session, aiohttp.ClientSession)
        assert not client.session.closed

    # Session should be closed after exiting context
    assert client.session is None


################################
# Authentication
################################


@pytest.mark.asyncio
async def test_check_auth() -> None:
    """Test authentication check."""
    # Initialize client
    client = WilmaClient("https://test.inschool.fi")

    # Not authenticated
    with pytest.raises(AuthenticationError):
        client._check_auth()

    # Set authentication values
    client.user_id = "!12345"
    client._sid = "test_sid"

    # Should not raise exception
    client._check_auth()


@pytest.mark.asyncio
@patch("wilhelmina.client.aiohttp.ClientSession")
async def test_login_success(mock_session) -> None:
    """Test successful login."""
    # Create client with patched session
    client = WilmaClient("https://test.inschool.fi", debug=False)

    # Login ID cookie
    login_id_cookie = MagicMock()
    login_id_cookie.value = "test_login_id"

    # Set up token response mock
    token_resp = AsyncMock()
    token_resp.status = 200
    token_resp.cookies = MagicMock()
    token_resp.cookies.get = MagicMock(return_value=login_id_cookie)

    # SID cookie
    sid_cookie = MagicMock()
    sid_cookie.value = "test_sid"

    # Set up login response mock
    login_resp = AsyncMock()
    login_resp.status = 303
    login_resp.cookies = MagicMock()
    login_resp.cookies.get = MagicMock(return_value=sid_cookie)
    login_resp.headers = {"Location": "https://test.inschool.fi/!12345?checkcookie"}

    # Configure session mock for get and post
    get_context = AsyncMock()
    get_context.__aenter__.return_value = token_resp
    post_context = AsyncMock()
    post_context.__aenter__.return_value = login_resp

    mock_session_instance = mock_session.return_value
    mock_session_instance.get.return_value = get_context
    mock_session_instance.post.return_value = post_context

    # Call login
    await client.login("testuser", "testpass")

    # Verify client state
    assert client.user_id == "!12345"
    assert client._sid == "test_sid"

    # Verify requests
    client.session.get.assert_called_once_with("https://test.inschool.fi/token")
    client.session.post.assert_called_once_with(
        "https://test.inschool.fi/login",
        data={"Login": "testuser", "Password": "testpass", "SESSIONID": "test_login_id"},
        allow_redirects=False,
    )


@pytest.mark.asyncio
async def test_login_failure() -> None:
    """Test login failure."""
    client = WilmaClient("https://test.inschool.fi", debug=False)

    # Create mock session
    mock_session = MagicMock()  # Use MagicMock instead of AsyncMock for the session

    # Mock token request response
    token_response = MagicMock()  # Use MagicMock instead of AsyncMock
    token_response.status = 200

    login_id_cookie = MagicMock()
    login_id_cookie.value = "test_login_id"

    cookies = MagicMock()
    cookies.get.side_effect = lambda name: {
        "Wilma2LoginID": login_id_cookie,
    }.get(name)

    token_response.cookies = cookies

    # Mock login response with failure status
    login_response = MagicMock()  # Use MagicMock instead of AsyncMock
    login_response.status = 401  # Unauthorized

    # Set up context manager returns
    mock_get_ctx = MagicMock()
    mock_get_ctx.__aenter__.return_value = token_response
    mock_session.get.return_value = mock_get_ctx

    mock_post_ctx = MagicMock()
    mock_post_ctx.__aenter__.return_value = login_response
    mock_session.post.return_value = mock_post_ctx

    # Patch _ensure_session to return our mock session
    with patch.object(client, "_ensure_session", AsyncMock(return_value=mock_session)):
        # Call login and expect exception
        with pytest.raises(AuthenticationError):
            await client.login("testuser", "wrong_password")


@pytest.mark.asyncio
@patch("wilhelmina.client.logger")
async def test_login_edge_cases(mock_logger) -> None:
    """Test edge cases for the login method."""
    # Test login with token failure
    client = WilmaClient("https://test.inschool.fi")

    # Create proper mocks using MagicMock instead of AsyncMock
    mock_session = MagicMock()

    # Mock the response for token request with error status
    token_response = MagicMock()
    token_response.status = 404  # Not found

    # Setup mock context manager that returns the token response
    mock_get_ctx = MagicMock()
    mock_get_ctx.__aenter__.return_value = token_response
    mock_session.get.return_value = mock_get_ctx

    # Set up _ensure_session to return our mock
    with patch.object(client, "_ensure_session", AsyncMock(return_value=mock_session)):
        # Should raise exception for token failure
        with pytest.raises(AuthenticationError, match="Failed to get token"):
            await client.login("testuser", "testpass")

        # Test missing login ID cookie
        token_response.status = 200  # Success
        token_response.cookies.get.return_value = None  # No cookie

        with pytest.raises(AuthenticationError, match="No login ID cookie found"):
            await client.login("testuser", "testpass")

        # Test login failure (status not 302/303)
        token_response.cookies.get.return_value = MagicMock(value="test_login_id")
        login_response = MagicMock()
        login_response.status = 401  # Unauthorized

        # Setup mock context manager for post
        mock_post_ctx = MagicMock()
        mock_post_ctx.__aenter__.return_value = login_response
        mock_session.post.return_value = mock_post_ctx

        with pytest.raises(AuthenticationError, match="Login failed"):
            await client.login("testuser", "testpass")


@pytest.mark.asyncio
async def test_login_response_errors() -> None:
    """Test various error conditions during login."""
    client = WilmaClient("https://test.inschool.fi")

    # Create and inject _ensure_session return value directly
    mock_session = MagicMock()

    # Configure the token response
    token_resp = MagicMock()
    token_resp.status = 200
    token_resp.cookies.get.return_value = MagicMock(value="test_login_id")

    # Configure the login response
    login_resp = MagicMock()
    login_resp.status = 302

    # Setup a custom cookies.get for login response that returns None for the first call
    # and a MagicMock for the second call
    cookies_get_mock = MagicMock()
    cookies_get_mock.side_effect = (
        lambda cookie_name: None if cookie_name == "Wilma2SID" else MagicMock(value="some_value")
    )
    login_resp.cookies.get = cookies_get_mock

    # Configure session get and post to return context managers with our responses
    mock_get_ctx = MagicMock()
    mock_get_ctx.__aenter__.return_value = token_resp
    mock_post_ctx = MagicMock()
    mock_post_ctx.__aenter__.return_value = login_resp

    mock_session.get.return_value = mock_get_ctx
    mock_session.post.return_value = mock_post_ctx

    # Mock _ensure_session to return our mock session
    with patch.object(client, "_ensure_session", AsyncMock(return_value=mock_session)):
        # Test case: Missing SID cookie
        with pytest.raises(AuthenticationError, match="No SID cookie found"):
            await client.login("testuser", "testpass")

        # Update the cookie mock to return a SID cookie
        cookies_get_mock.side_effect = lambda cookie_name: MagicMock(value="test_sid")

        # Test case: Missing Location header
        login_resp.headers = {}
        with pytest.raises(AuthenticationError, match="No Location header in response"):
            await client.login("testuser", "testpass")

        # Test case: Invalid Location header (without an exclamation mark)
        login_resp.headers = {"Location": "invalid_url_without_exclamation_mark"}

        with pytest.raises(AuthenticationError, match="Invalid Location header"):
            await client.login("testuser", "testpass")


@pytest.mark.asyncio
async def test_debug_mode() -> None:
    """Test client debug mode."""
    # Create client with debug=True
    WilmaClient("https://test.inschool.fi", debug=True)

    # Debug mode should set logger level to DEBUG
    from wilhelmina.client import logger

    assert logger.level == 10  # DEBUG level

    # Create client with debug=False
    WilmaClient("https://test.inschool.fi", debug=False)

    # Logger level should not be changed when debug=False
    assert logger.level == 10  # Still DEBUG from the previous client


################################
# Authenticated Requests
################################


@pytest.mark.asyncio
async def test_authenticated_request() -> None:
    """Test authenticated request method."""
    # Initialize client
    client = WilmaClient("https://test.inschool.fi")

    # Set authentication values
    client.user_id = "!12345"
    client._sid = "test_sid"

    # We need to patch _ensure_session to prevent actual HTTP requests
    with patch.object(client, "_ensure_session") as mock_ensure_session:
        # Create mock session
        mock_session = AsyncMock()
        mock_response = AsyncMock(spec=ClientResponse)
        mock_response.status = 200

        # Configure mock get method
        mock_session.get = AsyncMock(return_value=mock_response)

        # Make _ensure_session return our mock session
        mock_ensure_session.return_value = mock_session

        # Call the method
        await client._authenticated_request("path/to/resource")

        # Verify the request was made with the right URL and headers
        mock_session.get.assert_called_with(
            "https://test.inschool.fi/path/to/resource", headers={"Wilma2SID": "test_sid"}
        )

        # Test with template path
        mock_session.get.reset_mock()
        await client._authenticated_request("{user_id}/messages")

        # Verify URL was correctly formatted with user_id
        mock_session.get.assert_called_with(
            "https://test.inschool.fi/!12345/messages", headers={"Wilma2SID": "test_sid"}
        )

        # Test with params
        mock_session.get.reset_mock()
        params = {"format": "json", "id": 123}
        await client._authenticated_request("{user_id}/messages", params=params)

        mock_session.get.assert_called_with(
            "https://test.inschool.fi/!12345/messages",
            headers={"Wilma2SID": "test_sid"},
            params=params,
        )

        # Test POST request
        mock_session.post = AsyncMock(return_value=mock_response)
        data = {"key": "value"}
        await client._authenticated_request("{user_id}/messages", method="POST", data=data)

        mock_session.post.assert_called_with(
            "https://test.inschool.fi/!12345/messages", headers={"Wilma2SID": "test_sid"}, data=data
        )


@pytest.mark.asyncio
async def test_authenticated_request_http_error() -> None:
    """Test HTTP error handling in authenticated request."""
    client = WilmaClient("https://test.inschool.fi")
    client.user_id = "!12345"
    client._sid = "test_sid"

    # Setup mocks
    mock_session = AsyncMock()
    mock_response = AsyncMock()
    mock_response.status = 500  # Server error

    # Mock get method to return our response
    mock_session.get = AsyncMock(return_value=mock_response)

    # Setup client with mocks
    with patch.object(client, "_ensure_session", AsyncMock(return_value=mock_session)):
        # With status 500, should raise WilmaError containing the status code
        with pytest.raises(WilmaError) as exc_info:
            await client._authenticated_request("path/to/resource")

        # Verify error message contains status code
        assert "500" in str(exc_info.value)


################################
# Messages
################################


@pytest.mark.asyncio
@patch("wilhelmina.client.WilmaClient._get_unread_message_ids")
async def test_get_messages(mock_get_unread, authenticated_client, messages_data) -> None:
    """Test getting messages."""
    client = authenticated_client

    # Mock the _get_unread_message_ids method to return a set of IDs
    mock_get_unread.return_value = {1}  # Message ID 1 is unread

    # Set up response mock
    resp = AsyncMock()
    resp.status = 200
    resp.json.return_value = messages_data

    # Override _authenticated_request to return our mocked response
    client._authenticated_request = AsyncMock(return_value=resp)

    # Call get_messages
    messages = await client.get_messages()

    # Verify _authenticated_request was called
    client._authenticated_request.assert_called_once_with("{user_id}/messages/list")

    # Verify response
    assert len(messages) == 3
    assert isinstance(messages[0], Message)
    assert messages[0].id == 1
    assert messages[0].subject == "Test message 1"
    assert messages[0].timestamp == "2025-04-16 13:31"
    assert messages[0].sender == "Teacher 1"
    assert messages[0].unread is True  # First message should be unread based on our mock
    assert messages[1].unread is False  # Second message should be read


@pytest.mark.asyncio
@patch("wilhelmina.client.WilmaClient._get_unread_message_ids")
async def test_get_messages_with_filters(
    mock_get_unread, authenticated_client, messages_data, date_filters
) -> None:
    """Test getting messages with filters."""
    client = authenticated_client

    # Mock the _get_unread_message_ids method
    mock_get_unread.return_value = {1}  # Message ID 1 is unread

    # Set up response mock
    resp = AsyncMock()
    resp.status = 200
    resp.json.return_value = messages_data

    # Set the _authenticated_request method
    client._authenticated_request = AsyncMock(return_value=resp)

    # Test with only_unread=True
    messages = await client.get_messages(only_unread=True)
    assert len(messages) == 1
    assert messages[0].id == 1
    assert messages[0].unread is True

    # Test with after date filter (middle date)
    messages = await client.get_messages(after=date_filters["middle"])
    assert len(messages) == 2
    assert messages[0].id == 1  # Most recent
    assert messages[1].id == 2  # Middle

    # Test with after date - before all messages
    messages = await client.get_messages(after=date_filters["before_all"])
    assert len(messages) == 3  # All messages

    # Test with after date - after all messages
    messages = await client.get_messages(after=date_filters["after_all"])
    assert len(messages) == 0  # No messages

    # Test with timezone-aware date
    messages = await client.get_messages(after=date_filters["with_timezone"])
    assert len(messages) == 2  # Same as middle date test

    # Test with combination of filters
    messages = await client.get_messages(only_unread=True, after=date_filters["middle"])
    assert len(messages) == 1
    assert messages[0].id == 1
    assert messages[0].unread is True


@pytest.mark.asyncio
@patch("wilhelmina.client.WilmaClient._get_unread_message_ids")
@patch("wilhelmina.client.WilmaClient.get_message_content")
async def test_get_messages_with_content(
    mock_get_message_content,
    mock_get_unread,
    authenticated_client,
    messages_data,
    message_content_data,
) -> None:
    """Test getting messages with content."""
    client = authenticated_client

    # Mock the _get_unread_message_ids method
    mock_get_unread.return_value = {1}

    # Set up response mock for messages list
    resp = AsyncMock()
    resp.status = 200
    resp.json.return_value = messages_data
    client._authenticated_request = AsyncMock(return_value=resp)

    # Mock get_message_content to return messages with content
    full_message1 = Message.from_dict(message_content_data["messages"][0])
    full_message2 = Message.from_dict(
        {
            "Id": 2,
            "Subject": "Test message 2",
            "TimeStamp": "2025-04-15 10:20",
            "Folder": "Inbox",
            "SenderId": 456,
            "SenderType": 1,
            "Sender": "Teacher 2",
            "ContentHtml": "<p>Test message 2 full content.</p>",
            "AllowForward": True,
            "AllowReply": True,
            "ReplyList": [],
            "Senders": [{"Name": "Teacher 2", "Href": "/profiles/456"}],
        }
    )
    full_message3 = Message.from_dict(
        {
            "Id": 3,
            "Subject": "Test message 3",
            "TimeStamp": "2025-04-10 09:15",
            "Folder": "Inbox",
            "SenderId": 789,
            "SenderType": 1,
            "Sender": "Teacher 3",
            "ContentHtml": "<p>Test message 3 full content.</p>",
            "AllowForward": True,
            "AllowReply": True,
            "ReplyList": [],
            "Senders": [{"Name": "Teacher 3", "Href": "/profiles/789"}],
        }
    )

    # Configure mock to return different messages based on the message ID
    mock_get_message_content.side_effect = lambda id: (
        full_message1 if id == 1 else (full_message2 if id == 2 else full_message3)
    )

    # Call get_messages with with_content=True
    messages = await client.get_messages(with_content=True, no_message_content_fetch_limit=True)

    # Verify get_message_content was called for each message
    assert mock_get_message_content.call_count == 3

    # Verify the message content was updated
    assert len(messages) == 3
    assert messages[0].content_html == "<p>This is a test message content.</p>"
    assert messages[1].content_html == "<p>Test message 2 full content.</p>"
    assert messages[2].content_html == "<p>Test message 3 full content.</p>"


@pytest.mark.asyncio
@patch("wilhelmina.client.WilmaClient._get_unread_message_ids")
async def test_get_message_content(mock_get_unread, authenticated_client) -> None:
    """Test fetching message content."""
    client = authenticated_client

    # Mock _get_unread_message_ids
    mock_get_unread.return_value = {10}

    # Mock _authenticated_request
    mock_response = AsyncMock()
    mock_response.json = AsyncMock(
        return_value={
            "messages": [
                {
                    "Id": 10,
                    "Subject": "Test message",
                    "TimeStamp": "2025-04-16 13:31",
                    "Folder": "Inbox",
                    "SenderId": 123,
                    "SenderType": 1,
                    "Sender": "Teacher 1",
                    "ContentHtml": "<p>Message content</p>",
                    "AllowForward": True,
                    "AllowReply": True,
                    "ReplyList": [],
                    "Senders": [],
                }
            ]
        }
    )

    client._authenticated_request = AsyncMock(return_value=mock_response)

    # Call get_message_content
    message = await client.get_message_content(10)

    # Verify authenticated request was called correctly
    client._authenticated_request.assert_called_once_with(
        "{user_id}/messages/10", params={"format": "json"}
    )

    # Verify message properties
    assert message.id == 10
    assert message.subject == "Test message"
    assert message.content_html == "<p>Message content</p>"
    assert message.unread is True  # Should be in unread set


@pytest.mark.asyncio
@patch("wilhelmina.client.WilmaClient._get_unread_message_ids")
async def test_get_messages_with_invalid_content(mock_get_unread) -> None:
    """Test handling of invalid message content."""
    client = WilmaClient("https://test.inschool.fi")
    client.user_id = "!12345"
    client._sid = "test_sid"

    # Mock _get_unread_message_ids
    mock_get_unread.return_value = set()

    # Create complete message dict with all required fields
    message_dict = {
        "Id": 123,
        "Subject": "Test Subject",
        "TimeStamp": "2023-01-01 12:00",
        "Folder": "Inbox",
        "SenderId": 456,
        "SenderType": 1,
        "Sender": "Test Sender",
    }

    # Mock list response
    list_response = MagicMock()
    list_response.json = AsyncMock(return_value={"Messages": [message_dict]})

    # Mock _authenticated_request
    with (
        patch.object(client, "_authenticated_request", AsyncMock(return_value=list_response)),
        patch.object(
            client, "get_message_content", AsyncMock(side_effect=Exception("Failed to get content"))
        ),
    ):
        # Should handle exception and continue
        messages = await client.get_messages(with_content=True, no_message_content_fetch_limit=True)

        # Should have one message despite content fetch error
        assert len(messages) == 1
        assert messages[0].id == 123


################################
# Helper Functions
################################


@pytest.mark.asyncio
async def test_message_attribute_transfer() -> None:
    """Test transfer of attributes from one message to another."""
    # This tests line 305 in client.py

    from wilhelmina.models import Message

    # Create source message
    original_message = Message(
        id=123,
        subject="Original Subject",
        timestamp="2023-01-01 12:00",
        folder="Inbox",
        sender_id=456,
        sender_type=1,
        sender="Original Sender",
    )

    # Create message with additional attributes
    updated_message = Message(
        id=123,
        subject="Updated Subject",
        timestamp="2023-01-01 12:00",
        folder="Inbox",
        sender_id=456,
        sender_type=1,
        sender="Updated Sender",
        content_html="<p>New content</p>",  # New attribute not in original
    )

    # Simulate the code from line 305 that transfers attributes
    for attr, value in vars(updated_message).items():
        if attr != "unread" and getattr(original_message, attr, None) is None:
            setattr(original_message, attr, value)

    # The content_html attribute should have been copied
    assert original_message.content_html == "<p>New content</p>"
    # Original attributes should remain unchanged
    assert original_message.subject == "Original Subject"


def test_non_string_value_handling() -> None:
    """Test handling of a non-string value in message ID."""
    # Test both cases of the functionality
    # Case 1: Non-string value should be skipped
    value1 = 123  # Non-string value
    if not isinstance(value1, str):
        result1 = "skipped"
    else:
        result1 = "processed"

    # It should return skipped and not raise an exception
    assert result1 == "skipped"

    # Case 2: String value should be processed
    value2 = "456"  # String value
    if not isinstance(value2, str):
        result2 = "skipped"
    else:
        result2 = "processed"

    # It should return processed
    assert result2 == "processed"



"""Wilma API client."""

import importlib.util
import logging
import subprocess
import sys

from wilhelmina.client import AuthenticationError, WilmaClient, WilmaError
from wilhelmina.models import Message, MessagesList, Sender

logger = logging.getLogger(__name__)

__all__ = [
    "WilmaClient",
    "WilmaError",
    "AuthenticationError",
    "Message",
    "MessagesList",
    "Sender",
    "install_playwright_deps",
]


def install_playwright_deps(force: bool = False, quiet: bool = False) -> bool:
    """Install Playwright dependencies.

    Args:
        force: Force installation even if playwright is not installed
        quiet: Don't log anything

    Returns:
        True if installed, False if skipped or failed
    """
    if not force:
        # Check if playwright is installed
        if importlib.util.find_spec("playwright") is None:
            if not quiet:
                logger.warning(
                    "Playwright is not installed. Install with 'pip install wilma[playwright]'"
                )
            return False

    log_level = logging.ERROR if quiet else logging.INFO
    logger.setLevel(log_level)

    try:
        logger.info("Installing Playwright dependencies...")
        subprocess.run(
            [sys.executable, "-m", "playwright", "install", "chromium"],
            check=True,
            capture_output=quiet,
        )
        logger.info("Playwright dependencies installed successfully.")
        return True
    except (subprocess.SubprocessError, FileNotFoundError) as e:
        logger.exception("Failed to install Playwright dependencies: %s", e)
        return False


# Auto-install Playwright browser when package is installed with the extra
if importlib.util.find_spec("playwright") is not None:
    # Only attempt installation during import if playwright is available
    # This ensures the browser is installed automatically when the package
    # is installed with the playwright extra
    install_playwright_deps(quiet=True)



"""Tests for error handling in WilmaClient."""

from unittest.mock import AsyncMock, patch

import pytest

from wilhelmina.client import AuthenticationError, WilmaClient, WilmaError
from wilhelmina.models import Message


@pytest.mark.asyncio
async def test_not_authenticated() -> None:
    """Test that methods requiring authentication raise errors when not authenticated."""
    client = WilmaClient("https://test.inschool.fi")

    # Methods that require authentication
    with pytest.raises(AuthenticationError, match="Not authenticated"):
        await client._check_auth()

    with pytest.raises(AuthenticationError):
        await client.get_messages()

    with pytest.raises(AuthenticationError):
        await client.get_message_content(1)


@pytest.mark.asyncio
@patch("wilhelmina.client.WilmaClient._get_unread_message_ids")
async def test_message_content_fetch_limit(mock_get_unread, authenticated_client) -> None:
    """Test the message content fetch limit."""
    client = authenticated_client

    # Mock unread IDs
    mock_get_unread.return_value = set()

    # With more than 10 messages and no override, should fail
    # Build a response that has > 10 messages
    messages_with_many = {"Messages": []}
    for i in range(15):  # Add 15 messages
        messages_with_many["Messages"].append(
            {
                "Id": i + 1,
                "Subject": f"Test message {i + 1}",
                "TimeStamp": "2025-04-16 13:31",
                "Folder": "Inbox",
                "SenderId": 123,
                "SenderType": 1,
                "Sender": "Teacher 1",
                "ContentHtml": "<p>Message content</p>",
                "AllowForward": True,
                "AllowReply": True,
                "ReplyList": [],
                "Senders": [],
            }
        )

    # Set up response mock
    resp = AsyncMock()
    resp.status = 200
    resp.json.return_value = messages_with_many
    client._authenticated_request = AsyncMock(return_value=resp)

    # Try to fetch content for all messages without override - should fail
    with pytest.raises(WilmaError, match="Not allowed to fetch full message content"):
        await client.get_messages(with_content=True)


@pytest.mark.asyncio
@patch("wilhelmina.client.importlib.util")
@patch("wilhelmina.client._has_playwright", False)
async def test_get_unread_ids_no_playwright(mock_importlib, authenticated_client) -> None:
    """Test behavior when Playwright is not available."""
    client = authenticated_client

    # Mock importlib to make _has_playwright return False
    mock_importlib.find_spec.return_value = None

    # Call the method - should return empty set
    unread_ids = await client._get_unread_message_ids()

    # Should return empty set when Playwright is not available
    assert unread_ids == set()


@pytest.mark.asyncio
async def test_content_markdown_error() -> None:
    """Test error when accessing content_markdown with None content_html."""

    # Create message with None content_html
    message = Message(
        id=1,
        subject="Test message",
        timestamp="2025-04-16 13:31",
        folder="Inbox",
        sender_id=123,
        sender_type=1,
        sender="Teacher 1",
        content_html=None,
    )

    # Verify ValueError is raised when trying to access content_markdown
    with pytest.raises(ValueError, match="Message content is not available"):
        _ = message.content_markdown



"""Utility functions for Wilma client."""

import html2text

# Configure html2text
_html_converter = html2text.HTML2Text()
_html_converter.ignore_links = False
_html_converter.body_width = 0  # No wrapping
_html_converter.unicode_snob = True
_html_converter.protect_links = True
_html_converter.images_to_alt = True


def html_to_markdown(html: str) -> str:
    """Convert HTML to Markdown.

    Args:
        html: HTML content

    Returns:
        Markdown content
    """
    return _html_converter.handle(html).strip()
